from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from jwcrypto import jwk

# Generate an RSA private key
private_key = rsa.generate_private_key(
    public_exponent=65537, key_size=2048, backend=default_backend()
)

# Serialize the private key to PEM format
private_pem = private_key.private_bytes(
    encoding=serialization.Encoding.PEM,
    format=serialization.PrivateFormat.PKCS8,
    encryption_algorithm=serialization.NoEncryption(),
)

# Write the private key PEM file to disk
with open("private_key.pem", "wb") as pem_file:
    pem_file.write(private_pem)

# Extract the public key from the private key
public_key = private_key.public_key()

# Serialize the public key to PEM format
public_pem = public_key.public_bytes(
    encoding=serialization.Encoding.PEM,
    format=serialization.PublicFormat.SubjectPublicKeyInfo,
)

# Write the public key PEM file to disk
with open("public_key.pem", "wb") as pem_file:
    pem_file.write(public_pem)

print("The private key has been generated and saved to 'private_key.pem'")
print("The public key has been generated and saved to 'public_key.pem'")


# Load the private key from the PEM file
with open("private_key.pem", "rb") as pem_file:
    private_key_pem = pem_file.read()

# Load the public key from the PEM file
with open("public_key.pem", "rb") as pem_file:
    public_key_pem = pem_file.read()

# Convert the private key PEM to a JWK object
private_key = jwk.JWK.from_pem(private_key_pem)

# Convert the public key PEM to a JWK object
public_key = jwk.JWK.from_pem(public_key_pem)

# Create a JWKS with both the private and public keys
jwks = {
    "keys": [
        private_key.export(private_key=True, as_dict=True),
        public_key.export(as_dict=True),
    ]
}

# Convert JWKS to a JSON string
import json

jwks_json = json.dumps(jwks, indent=4)

# Print the JWKS JSON
print(jwks_json)

# Optionally, save the JWKS to a file
with open("jwks.json", "w") as jwks_file:
    jwks_file.write(jwks_json)

print("The JWKS has been generated and saved to 'jwks.json'")



from typing import Any, Optional, Self

import jwt as pyjwt
from cachetools.func import ttl_cache
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from fastapi import HTTPException, Request
from httpx import Client, HTTPError
from jwt import (
    InvalidSignatureError,
    InvalidTokenError,
    MissingRequiredClaimError,
    PyJWKClient,
    PyJWKSet,
)
from loguru import logger
from pydantic import BaseModel
from redbaby.pyobjectid import PyObjectId

from ..authproviders.models import AuthProviderDAO
from ..entities.models import EntityDAO
from ..entities.schemas import EntityIntermediate
from ..schemas import Creator, Infostar
from ..utils import creation, reading


def get_token_headers(token: str) -> dict[str, Any]:
    header = pyjwt.get_unverified_header(token)
    return header


def get_token_unverified_claims(token: str) -> dict[str, Any]:
    claims = pyjwt.decode(token, options={"verify_signature": False})
    return claims


def get_signing_key(kid: str, domain: str) -> Optional[str]:
    jwk_set = ManyJSONKeySetStore.get_jwks(domain)
    signing_key = PyJWKClient.match_kid(jwk_set.keys, kid)
    if isinstance(signing_key, RSAPrivateKey):
        return signing_key.key.public_key()
    return signing_key.key if signing_key else None


class Auth0Settings(BaseModel):
    domain: str
    audience: str

    @classmethod
    def from_authprovider(cls, authprovider: AuthProviderDAO) -> Self:
        iss = authprovider.get_external_id("issuer")
        if not iss:
            raise MissingRequiredClaimError("iss")
        aud = authprovider.get_external_id("audience")
        if not aud:
            raise MissingRequiredClaimError("aud")
        return cls(domain=iss, audience=aud)


class ManyJSONKeySetStore:
    @classmethod
    @ttl_cache(maxsize=16, ttl=60 * 60 * 6)
    def get_jwks(cls, domain: str) -> PyJWKSet:
        logger.debug(f"Fetching JWKS from {domain}.")
        with Client() as client:
            res = client.get(f"{domain}.well-known/jwks.json")
        try:
            res.raise_for_status()
        except HTTPError as e:
            logger.error(f"Failed to fetch JWKS from {domain}.")
            raise e

        return PyJWKSet.from_dict(res.json())


class RequestAuthenticator:
    @staticmethod
    def get_authprovider(token_value: str) -> AuthProviderDAO:
        logger.debug("Getting AuthProvider.")
        token_claims = get_token_unverified_claims(token_value)
        filters: dict[str, Any] = {"type": "auth0"}
        if aud := token_claims.get("aud"):
            # We assume that the actual audience is the first element in the list.
            # The second element is the issuer's "userinfo" endpoint.
            filters["external_ids"] = {
                "$elemMatch": {"name": "audience", "value": aud[0]}
            }

        provider = reading.read_one_filters(infostar={}, model=AuthProviderDAO, **filters)  # type: ignore
        return provider

    @staticmethod
    def validate_access_token(
        token_value: str,
        token_headers: dict[str, Any],
        authprovider: AuthProviderDAO,
    ) -> dict:
        logger.debug("Validating access token.")
        sets = Auth0Settings.from_authprovider(authprovider)
        kid = token_headers.get("kid")
        if kid is None:
            raise InvalidTokenError("Missing 'kid' header.")

        signing_key = get_signing_key(kid, sets.domain)
        if signing_key is None:
            raise InvalidSignatureError("No signing key found.")

        access_claims = pyjwt.decode(
            token_value,
            signing_key,
            algorithms=[token_headers.get("alg", "RS256")],
            issuer=sets.domain,
            audience=sets.audience,
            options={"require": ["exp", "iss", "aud"]},
        )
        return access_claims

    @staticmethod
    def validate_id_token(
        token_value: str,
        token_headers: dict[str, Any],
        authprovider: AuthProviderDAO,
    ) -> dict:
        logger.debug("Validating ID token.")
        sets = Auth0Settings.from_authprovider(authprovider)
        kid = token_headers.get("kid")
        if kid is None:
            raise InvalidTokenError("Missing 'kid' header.")

        signing_key = get_signing_key(kid, sets.domain)
        if signing_key is None:
            raise InvalidSignatureError("No signing key found.")

        id_claims = pyjwt.decode(
            token_value,
            signing_key,
            algorithms=[token_headers.get("alg", "RS256")],
            issuer=sets.domain,
            options={"require": ["iss", "exp"], "verify_aud": False},
        )
        return id_claims

    @staticmethod
    def assemble_user_data(access_claims, id_claims) -> dict:
        logger.debug("Assembling user data.")
        required_access = ["org_id"]
        required_id = ["sub", "email"]
        for required_claims, claims in zip(
            [required_access, required_id], (access_claims, id_claims)
        ):
            for c in required_claims:
                if c not in claims:
                    raise MissingRequiredClaimError(c)

        user_data = {
            "user_id": id_claims.get("sub"),
            "user_email": id_claims.get("email"),
            "org_id": access_claims.get("org_id"),
        }
        return user_data

    @staticmethod
    def assemble_creator(infostar: Infostar) -> Creator:
        logger.debug("Assembling Creator based on Infostar.")
        c = Creator(
            client_name=f"{infostar.user_owner_handle}/{infostar.service_handle}",
            token_name=infostar.apikey_name,
            user_email=infostar.user_handle,
            user_ip=infostar.client_ip,
        )
        return c

    @staticmethod
    def assemble_infostar(
        request: Request, user_data: dict, authprovider: AuthProviderDAO
    ) -> Infostar:
        logger.debug("Assembling Infostar.")
        if request.client is not None:
            ip = request.client.host
        elif request.headers.get("x-tauth-ip"):
            ip = request.headers["x-tauth-ip"]
        elif request.headers.get("x-forwarded-for"):
            ip = request.headers["x-forwarded-for"]
        else:
            raise HTTPException(
                500,
                detail="Client's IP was not found in: request.client.host, X-Tauth-IP, X-Forwarded-For.",
            )

        infostar = Infostar(
            request_id=PyObjectId(),
            apikey_name="jwt",
            authprovider_type=authprovider.type,
            authprovider_org=authprovider.organization_ref.handle,
            extra={},
            service_handle=authprovider.service_ref.handle,  # TODO: azp
            user_handle=user_data["user_email"],
            user_owner_handle=user_data["org_id"],  # TODO map this to an org
            client_ip=ip,
            original=None,
        )

        try:
            filters = {
                "type": "user", 
                "handle": user_data["user_email"], 
                "owner_ref.handle": authprovider.organization_ref.handle,
            }
            reading.read_one_filters(
                infostar=infostar,
                model=EntityDAO,
                **filters
            )
        except HTTPException as e:
            if e.status_code in (404, 409):
                user_i = EntityIntermediate(
                    handle=user_data["user_email"],
                    owner_ref=authprovider.organization_ref.model_dump(),  # type: ignore
                    type="user",
                )
                creation.create_one(user_i, EntityDAO, infostar=infostar)
            else:
                raise e

        return infostar

    @classmethod
    def validate(cls, request: Request, token_value: str, id_token: str):
        try:
            header = get_token_headers(token_value)
            authprovider = cls.get_authprovider(token_value)
            access_claims = cls.validate_access_token(token_value, header, authprovider)
            id_claims = cls.validate_id_token(id_token, header, authprovider)
        except (
            MissingRequiredClaimError,
            InvalidTokenError,
            InvalidSignatureError,
            HTTPError,
        ) as e:
            raise HTTPException(
                401,
                detail={
                    "loc": ["headers", "Authorization"],
                    "msg": f"{e.__class__.__name__}: {e}",
                    "type": e.__class__.__name__,
                },
            )

        user_data = cls.assemble_user_data(access_claims, id_claims)
        request.state.infostar = cls.assemble_infostar(request, user_data, authprovider)
        request.state.creator = cls.assemble_creator(request.state.infostar)



import dotenv
import pytest
from fastapi.testclient import TestClient
from pymongo import MongoClient
from pymongo.database import Database

from tauth.app import create_app
from tauth.settings import Settings

from .utils import validate_isostring, validate_nonempty_string, validate_token

dotenv.load_dotenv()


@pytest.fixture(scope="session")
def mongo_client() -> MongoClient:
    client = MongoClient(Settings.get().TAUTH_MONGODB_URI)
    return client


@pytest.fixture(scope="session")
def tauth_db(mongo_client: MongoClient) -> Database:
    return mongo_client[Settings.get().TAUTH_MONGODB_DBNAME]


@pytest.fixture(scope="session")
def client() -> TestClient:
    return TestClient(create_app())


@pytest.fixture(scope="session")
def test_token_value() -> str:
    return Settings().TAUTH_ROOT_API_KEY


@pytest.fixture(scope="session")
def user_email() -> str:
    return "user@org.com"


@pytest.fixture(scope="session")
def headers(test_token_value: str, user_email: str) -> dict:
    obj = {
        "Authorization": f"Bearer {test_token_value}",
        "X-User-Email": user_email,
    }
    return obj


@pytest.fixture(scope="session")
def client_obj() -> dict:
    obj = {"name": "/example_app"}
    return obj


@pytest.fixture(scope="session")
def expectations_creator():
    exp = {
        "client_name": validate_nonempty_string,
        "user_email": validate_nonempty_string,
        "token_name": validate_nonempty_string,
    }
    return exp


@pytest.fixture(scope="session")
def expectations_token_creation_obj(expectations_creator):
    validations = {
        "created_at": validate_isostring,
        "created_by": expectations_creator,
        "name": "default",
        "value": validate_token,
    }
    return validations


@pytest.fixture(scope="session")
def access_token() -> str:
    return "eyJhbGciOiJSUzI1NiIsImtpZCI6IlFodUdSLWZ5aWttVGVOaEVCYzRDYnozUVhrOHdNWm10R0NPQjNPcHlMVm8iLCJ0eXAiOiJKV1QifQ.eyJpc3MiOiJodHRwczovL3Rlc3QvIiwiYXVkIjoiaHR0cHM6Ly90ZXN0LyIsImV4cCI6MjE0NzQ4MzY0Nywib3JnX2lkIjoidGVzdCIsInN1YiI6InRlc3QiLCJlbWFpbCI6InRlc3RAZ21haWwuY29tIn0.SkhvwwVibrFktqeISs0soMI0gFzJWnN2RqtLiNSYxEydGacwNimKZ9GTiz01m0gVzKbcTt4hV567ohCCvVU969FFRD-9UmtoTvG9gSE6WnJORryKq5UrNKAjU6TP07Y3zy2FOVOCrO-4y7c04rCAPA6oeL2JxLpf0n9t2D1rzsX7YpdIGWxyFTNcaxUwzx3x4f0NXtQuu-HN07OumA5dPa68Xq8zAheOFYx91O_RTRYXr236KKy7oytoGdl7yFz-p4nICycuG-RllYcOCwxHUxN3m77DAgjeiIdHLulpL6FkSmMFTdXr-cwkddBIlCwy3nMdQDbBv3NBlu6V-kcwbw"


@pytest.fixture(scope="session")
def id_token() -> str:
    return "eyJhbGciOiJSUzI1NiIsImtpZCI6IlFodUdSLWZ5aWttVGVOaEVCYzRDYnozUVhrOHdNWm10R0NPQjNPcHlMVm8iLCJ0eXAiOiJKV1QifQ.eyJpc3MiOiJodHRwczovL3Rlc3QvIiwiYXVkIjoiaHR0cHM6Ly90ZXN0LyIsImV4cCI6MjE0NzQ4MzY0Nywib3JnX2lkIjoidGVzdCIsInN1YiI6InRlc3QiLCJlbWFpbCI6InRlc3RAZ21haWwuY29tIn0.SkhvwwVibrFktqeISs0soMI0gFzJWnN2RqtLiNSYxEydGacwNimKZ9GTiz01m0gVzKbcTt4hV567ohCCvVU969FFRD-9UmtoTvG9gSE6WnJORryKq5UrNKAjU6TP07Y3zy2FOVOCrO-4y7c04rCAPA6oeL2JxLpf0n9t2D1rzsX7YpdIGWxyFTNcaxUwzx3x4f0NXtQuu-HN07OumA5dPa68Xq8zAheOFYx91O_RTRYXr236KKy7oytoGdl7yFz-p4nICycuG-RllYcOCwxHUxN3m77DAgjeiIdHLulpL6FkSmMFTdXr-cwkddBIlCwy3nMdQDbBv3NBlu6V-kcwbw"


@pytest.fixture(scope="session")
def jwk() -> dict:
    return {
        "keys": [
            {
                "kty": "RSA",
                "n": "5J0I_Gidm_l6s4gxjSHLMsc9dJGnvOgLaE3CrmP-iftv8JWvgG9SBCB1pPtJgyOyh4ZQHIjG2wbePQZCgynRmEPFTlT_LGI84iBx6aMONGR_JdAnQv8Kw8f2ULP9vGgZLIr2OlUvX9FrmMuYGld7N21ABNH6CmtoWMSp71OcK_BhAT2SvZPpnUkblGLsOt_qPngyElFowqPuO-fozL2SPZYn2SiD9IoDr-upw-f9pGWI-7juJtDX0v3q6v7PKA44CunhWwZifiKhEBmLXc68ZJmRbLuRnXR3v22gl30ApYRWguZCT9o8lEZjsBZzNhJr4zCWGfK1-RYd4Cs7LwLnHQ",
                "e": "AQAB",
                "d": "UlaMyfwT5_1uyAKhdidZvCwuYjGjrFIW6bY4C_9PyfqZUplW4Hc3nuzh8k3fKDBPOKiTafOJS5GpsWjzw7HoH2MpSREr5pxrOTZeULu9fflEIiZaPbMF-YWnnWF2XclQZ86U6GWN7oDKs_mACty_MDNU_2dxGxecOXx7iAshEGGRBOGQ7j8SOiUE4yMjuqCqOOPqENOFvuzMo-cVUsc84U8D9UwX9yc3CvyTqYSRk6-BePKr_Az7lbbOcjM9TGKFePQIwdYSaqyk3zcm3mLGDAePxvZOjX6V978HBktFRmxdIU_Df6mG8TO1lA_mRM6IT_0dh1n-bnGEJcP4SaqdMQ",
                "p": "9f_FUEUVpxn7tenD_rrxlQzeNEH8WJs1y93zfmUXPho68WHFEVWe31_AHp9aLQwYx9oCu5S-HVRgPchUFjVMB_a2_jLeIMbzfucwcx3O5Gxa_nbzkkckeklAOGfZgTkjugdyWpbS4zZjgD8oBpOBIuP0oBBUVhXj6h_HCzq5Te8",
                "q": "7ehSwe4wmZyaPO27s95e0HhKsxnkSL3aYNSFpZoN18xJ7suMxw0ecIMmDHF39bpBreiIM7ZfsWX9-9IAJeLAmxvyqaegrwgEDz8fQcAXDG6eJdNmU7feB7fN5GAzv8jZ3sCjW1fiW-WO05DNKwh22mNCPC6V8GPIRwVpXvaOJ7M",
                "dp": "KHLd5sz1OnyzPr4pVAE5J-DtyHnxHECpH9Rm7SmCINv_RSFmXetOLDx6Qo7BLRcIHBRkzqMuf3QYPxBpgx1QWx7eB_4lOA5-iydIeCU2l6iZba3xbuPzw4e3345z3SOgFD6VNwAFvQZ8ZeH8mtg2K55_4rHMrDr9Nsny2I3XWlE",
                "dq": "l0Ruykvurt053J-0B0vOsXKq3HIMCoxce9DlIURwpNOJ9sGJw05-Gj_pQ2QqSB2jgTYm0-qTsbIN6b2-xlgE6xq8Ek85BdFs81zJPP8sKIV1HMvrdjOkgvfFe_4HKVxF2zJVK9EXZiLxy2d4bHI3T-hoxudAAtcPEslIFE9kV60",
                "qi": "3LjVAz8GJt_AXERcO7_850PCNu9CiTvTcua5CGvBV43B1nv97DAvI9_QBRyjeiBjT6SZZn06FUFiDQYfSYVwOdfQ9YnIub_zOJAbPOCD9cd-rdISHhnDsZtowKN-TOmAIGX5bykuKaVFaNExyZK5E6bDFBJPg7tm0RdoISHsByU",
                "kid": "QhuGR-fyikmTeNhEBc4Cbz3QXk8wMZmtGCOB3OpyLVo",
            },
            {
                "kty": "RSA",
                "n": "5J0I_Gidm_l6s4gxjSHLMsc9dJGnvOgLaE3CrmP-iftv8JWvgG9SBCB1pPtJgyOyh4ZQHIjG2wbePQZCgynRmEPFTlT_LGI84iBx6aMONGR_JdAnQv8Kw8f2ULP9vGgZLIr2OlUvX9FrmMuYGld7N21ABNH6CmtoWMSp71OcK_BhAT2SvZPpnUkblGLsOt_qPngyElFowqPuO-fozL2SPZYn2SiD9IoDr-upw-f9pGWI-7juJtDX0v3q6v7PKA44CunhWwZifiKhEBmLXc68ZJmRbLuRnXR3v22gl30ApYRWguZCT9o8lEZjsBZzNhJr4zCWGfK1-RYd4Cs7LwLnHQ",
                "e": "AQAB",
                "kid": "QhuGR-fyikmTeNhEBc4Cbz3QXk8wMZmtGCOB3OpyLVo",
            },
        ]
    }



import pytest
from fastapi.testclient import TestClient
from pymongo.database import Database

from .utils import run_validations


@pytest.fixture(autouse=True)
def teardown(tauth_db: Database):
    try:
        yield None
    finally:
        tauth_db["entities"].delete_many({"name": "/example_app"})
        tauth_db["melt_keys"].delete_many({"name": {"$in": ["default", "second"]}})


@pytest.fixture(scope="module")
def token_creation_request_obj() -> dict:
    obj = {"name": "second"}
    return obj


@pytest.mark.order(after="test_clients.py::test_create_one")
def test_create_one(
    client: TestClient, headers: dict, client_obj: dict, token_creation_request_obj: dict, expectations_token_creation_obj: dict
) -> None:
    _ = client.post("/api/tauth/clients/", json=client_obj, headers=headers)
    response = client.post(f"/api/tauth/clients/{client_obj['name']}/tokens", json=token_creation_request_obj, headers=headers)
    assert response.status_code == 201
    response_body = response.json()
    expectations = dict(**expectations_token_creation_obj)
    expectations["name"] = token_creation_request_obj["name"]
    run_validations(response_body, expectations)
    response2 = client.post(f"/api/tauth/clients/{client_obj['name']}/tokens", json=token_creation_request_obj, headers=headers)
    assert response2.status_code == 409
    # TODO: check error body



from datetime import datetime

from pydantic import BaseModel, Field
from redbaby.pyobjectid import PyObjectId

from .infostar import Infostar


class GeneratedFields(BaseModel):
    id: str | PyObjectId = Field(alias="_id")
    created_by: Infostar
    created_at: datetime



from typing import Optional

from loguru import logger
from opa_client import OpaClient
from opa_client.errors import ConnectionsError, DeletePolicyError, RegoParseError

from ...entities.models import EntityDAO
from ...settings import Settings
from ..policies.models import AuthorizationPolicyDAO
from .interface import AuthorizationInterface


class OPAEngine(AuthorizationInterface):
    def __init__(self):
        logger.debug("Attempting to establish connection with OPA Engine.")
        self.client = OpaClient()
        try:
            self.client.check_connection()
        except ConnectionsError as e:
            logger.error(f"Failed to establish connection with OPA: {e}")
            raise e
        logger.debug("OPA Engine is running.")

    def is_authorized(
        self,
        entity: EntityDAO,
        policy_name: str,
        resource: str,
        context: Optional[dict] = None,
    ) -> bool:
        opa_context = dict(input={})
        entity_json = entity.model_dump(mode="json")
        opa_context["input"]["entity"] = entity_json
        if context:
            opa_context["input"] |= context
        opa_result = self.client.check_permission(
            input_data=opa_context,
            policy_name=policy_name,
            rule_name=resource,
        )
        logger.debug(f"Raw OPA result: {opa_result}")
        opa_result = opa_result["result"]
        return opa_result

    def get_filters(
        self,
        entity: EntityDAO,
        policy_name: str,
        resource: str,
        context: Optional[dict] = None,
    ) -> dict:
        opa_context = dict(input=dict(context=context, entity=entity))
        opa_result = self.client.check_permission(
            input_data=opa_context,
            policy_name=policy_name,
            rule_name=resource,
        )
        logger.debug(f"Raw OPA result: {opa_result}")
        return opa_result

    def upsert_policy(self, policy_name: str, policy_content: str) -> bool:
        logger.debug(f"Upserting policy {policy_name!r} in OPA.")
        try:
            result = self.client.update_opa_policy_fromstring(
                policy_content,
                policy_name,
            )
        except RegoParseError as e:
            logger.error(f"Failed to upsert policy in OPA: {e}")
            raise e

        if not result:
            logger.error(f"Failed to upsert policy in OPA: {result}")
        return result

    def delete_policy(self, policy_name: str) -> bool:
        logger.debug(f"Deleting policy: {policy_name}.")
        try:
            result = self.client.delete_opa_policy(policy_name)
        except DeletePolicyError as e:
            logger.error(f"Failed to delete policy in OPA: {e}")
            raise e

        if not result:
            logger.error(f"Failed to upsert policy in OPA: {result}")
        return result



from authlib.jose import jwt

max_timestamp = 2**31 - 1

header = {"alg": "RS256"}
payload = {
    "iss": "https://test/",
    "aud": "test",
    "exp": max_timestamp,
}
with open("private_key.pem") as f:
    private_key = f.read()

s = jwt.encode(header, payload, private_key)

with open("public_key.pem") as f:
    public_key = f.read()

claims = jwt.decode(s, public_key)
print("Token:", s)
print("Claims:", claims)
try:
    claims.validate()
    is_valid = True
except:
    is_valid = False

print("Is valid:", is_valid)



"""
FastAPI dependency injection for privilege checking.

is_valid_user - anyone with a valid key
is_valid_admin - token_name == default
is_valid_superuser - token_name == default and client_name == $root
"""

from typing import Callable

from fastapi import HTTPException, Request, status

from ..schemas import Infostar


def get_infostar(
    validate_access: Callable[[Infostar], bool]
) -> Callable[[Request], Infostar]:
    def wrapper(request: Request) -> Infostar:
        infostar: Infostar = request.state.infostar
        if validate_access(infostar):
            return infostar
        else:
            s = status.HTTP_403_FORBIDDEN
            d = {  # TODO: use http_error_schemas
                "msg": "You do not have access to this resource.",
                "info": {"infostar": infostar.model_dump(mode="json")},
            }
            # TODO: delegate exception raising to the wrapped function
            raise HTTPException(status_code=s, detail=d)

    return wrapper


@get_infostar
def is_valid_user(infostar: Infostar) -> bool:
    return True


@get_infostar
def is_valid_admin(infostar: Infostar) -> bool:
    return infostar.apikey_name == "default"


@get_infostar
def is_valid_superuser(infostar: Infostar) -> bool:
    return infostar.apikey_name == "default" and infostar.authprovider_org == "/"


@get_infostar
def is_direct_user(infostar: Infostar) -> bool:
    return infostar.apikey_name != "jwt"



import pytest
from fastapi.testclient import TestClient
from pymongo.database import Database

from .utils import run_validations, validate_id, validate_isostring


@pytest.fixture(autouse=True)
def teardown(tauth_db: Database):
    try:
        yield None
    finally:
        tauth_db["entites"].delete_many({"name": "/example_app"})
        tauth_db["melt_keys"].delete_many({"name": {"$in": ["default", "second"]}})


@pytest.fixture(scope="module")
def expectations_token_metadata_obj(expectations_creator) -> dict:
    validations = {
        "created_at": validate_isostring,
        "created_by": expectations_creator,
        "name": lambda k, obj: isinstance(obj[k], str) and len(obj[k]) > 0,
    }
    return validations


@pytest.fixture(scope="module")
def expectations_client_creation_obj(expectations_token_creation_obj) -> dict:
    obj = {
        "created_at": validate_isostring,
        "_id": validate_id,
        "name": "/example_app",
        "tokens": [expectations_token_creation_obj],
        "users": lambda k, obj: isinstance(obj[k], list) and len(obj[k]) == 0,
    }
    return obj


@pytest.fixture(scope="module")
def expectations_client_view_obj(expectations_token_metadata_obj) -> dict:
    obj = {
        "created_at": validate_isostring,
        "_id": validate_id,
        "name": "/example_app",
        "tokens": [expectations_token_metadata_obj],
    }
    return obj


def test_create_one(
    client: TestClient,
    headers: dict,
    client_obj: dict,
    expectations_client_creation_obj: dict,
) -> None:
    response = client.post("/api/tauth/clients", json=client_obj, headers=headers)
    assert response.status_code == 201
    response_body = response.json()
    run_validations(response_body, expectations_client_creation_obj)


@pytest.mark.order(after="test_create_one")
def test_read_one(
    client: TestClient,
    headers: dict,
    client_obj: dict,
    expectations_client_view_obj: dict,
) -> None:
    _ = client.post("/api/tauth/clients", json=client_obj, headers=headers)
    response = client.get(f"/api/tauth/clients/{client_obj['name']}", headers=headers)
    assert response.status_code == 200
    response_body = response.json()
    run_validations(response_body, expectations_client_view_obj)



import pytest
from fastapi.testclient import TestClient
from pymongo.database import Database

from ..utils import run_validations


def non_leading_slash_client_name_error_message(value):
    return isinstance(value, str) and "absolute" in value and "slash" in value


def has_standard_error_fields(obj: dict):
    assert "detail" in obj
    inner_obj = obj["detail"]
    keys = ["loc", "msg", "type"]
    for k in keys:
        assert k in inner_obj


def begins_with_body_and_name(obj):
    assert obj["detail"]["loc"][0] == "body"
    assert obj["detail"]["loc"][1] == "name"


def is_non_empty_string(obj):
    assert isinstance(obj, str) and len(obj) > 0


def test_non_leading_slash_client_name(client: TestClient, headers: dict):
    response = client.post("/api/tauth/clients/", json={"name": "example_app"}, headers=headers)
    assert response.status_code == 422
    response_body = response.json()
    has_standard_error_fields(response_body)
    details_obj = response_body["detail"]
    begins_with_body_and_name(response_body)
    assert non_leading_slash_client_name_error_message(details_obj["msg"])
    assert details_obj["type"] == "InvalidClientName"


def test_invalid_client_names(client: TestClient, headers: dict):
    invalid_client_names = ["/example--app", "//example_app", "/example app", "/example\napp", "/example\tapp"]
    for name in invalid_client_names:
        response = client.post("/api/tauth/clients/", json={"name": name}, headers=headers)
        assert response.status_code == 422
        response_body = response.json()
        has_standard_error_fields(response_body)
        begins_with_body_and_name(response_body)
        is_non_empty_string(response_body["detail"]["msg"])
        assert response_body["detail"]["type"] == "InvalidClientName"



from pathlib import Path
from typing import Optional, cast

from fastapi import APIRouter, Body, Depends, HTTPException, Query, Request
from fastapi import status as s

from ..authz import privileges
from ..entities.models import EntityDAO, OrganizationRef, ServiceRef
from ..schemas import Infostar
from ..schemas.gen_fields import GeneratedFields
from ..utils import creation, reading
from .models import AuthProviderDAO
from .schemas import AuthProviderIn, AuthProviderMoreIn

service_name = Path(__file__).parent.name
router = APIRouter(prefix=f"/{service_name}", tags=[service_name])


@router.post("", status_code=s.HTTP_201_CREATED)
@router.post("/", status_code=s.HTTP_201_CREATED, include_in_schema=False)
async def create_one(
    request: Request,
    body: AuthProviderIn = Body(
        openapi_examples=AuthProviderIn.model_config["json_schema_extra"]["examples"][0]  # type: ignore
    ),
    infostar: Infostar = Depends(privileges.is_valid_admin),
):
    if body.service_name:
        service_ref = EntityDAO.from_handle_to_ref(body.service_name)
        if not service_ref:
            raise HTTPException(
                status_code=s.HTTP_404_NOT_FOUND,
                detail=f"Provided service name {body.service_name!r} not found.",
            )
        service_ref = cast(ServiceRef, service_ref)
        service_ref = ServiceRef(**service_ref.model_dump())
    else:
        service_ref = None
    org_ref = EntityDAO.from_handle_to_ref(body.organization_name)
    if org_ref is None:
        raise ValueError(f"Organization {body.organization_name} not found.")
    org_ref = OrganizationRef(**org_ref.model_dump())
    in_schema = AuthProviderMoreIn(
        **body.model_dump(),
        service_ref=service_ref,
        organization_ref=org_ref,
    )
    org = creation.create_one(in_schema, AuthProviderDAO, infostar)
    return GeneratedFields(**org.model_dump(by_alias=True))


@router.get("", status_code=s.HTTP_200_OK)
@router.get("/", status_code=s.HTTP_200_OK, include_in_schema=False)
async def read_many(
    request: Request,
    infostar: Infostar = Depends(privileges.is_valid_user),
    name: Optional[str] = Query(None),
):
    orgs = reading.read_many(
        infostar=infostar,
        model=AuthProviderDAO,
        **request.query_params,
    )
    return orgs



"""
Migration script to convert data from v1 to v2.
"""

import argparse
from collections import defaultdict
from typing import Callable, Literal

from loguru import logger
from pymongo import MongoClient, errors
from pymongo.collection import Collection
from pymongo.database import Database
from redbaby.pyobjectid import PyObjectId
from tqdm import tqdm

from tauth.authproviders.models import AuthProviderDAO
from tauth.entities.models import EntityDAO, EntityRef, OrganizationRef, ServiceRef
from tauth.legacy.tokens import TokenDAO
from tauth.schemas import Infostar
from tauth.schemas.attribute import Attribute

OID = PyObjectId()
MY_IP = "127.0.0.1"
MY_EMAIL = "sysadmin@teialabs.com"

SYSTEM_INFOSTAR = Infostar(
    request_id=OID,
    apikey_name="migrations",
    authprovider_org="/",
    authprovider_type="melt-key",
    extra={},
    service_handle="tauth",
    user_handle=MY_EMAIL,
    user_owner_handle="/",
    original=None,
    client_ip=MY_IP,
)


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--host",
        type=str,
        default="mongodb://localhost:27017",
        help="MongoDB host.",
    )
    parser.add_argument("--port", type=int, default="27017", help="MongoDB port.")
    parser.add_argument(
        "--in-db", type=str, default="tauth", help="MongoDB input database."
    )
    parser.add_argument(
        "--out-db", type=str, default="tauth-v2", help="MongoDB output database."
    )
    parser.add_argument("--dry-run", action="store_true", help="Dry run.")
    args = parser.parse_args()
    return args


class Partial:
    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __call__(self, *args, **kwargs):
        return self.func(*self.args, **self.kwargs, **kwargs)

    def __repr__(self) -> str:
        return f"Partial({self.func!r}, {self.args!r}, {self.kwargs!r})"

    def __str__(self) -> str:
        return f"Partial<{self.func}({self.args}, {self.kwargs})>"


def conditional_action(
    action: Partial, dry_run: bool, inaction: Callable[[Partial], None] = logger.info
):
    if not dry_run:
        logger.debug(f"Executing action: {action}")
        action()
    else:
        logger.debug(f"Would execute: {action}")
        inaction(action)


def conditional_insert(collection: Collection, document: dict, dry_run: bool):
    if not dry_run:
        logger.info(f"Inserting document into {collection.name}: {document}")
        collection.insert_one(document)
    else:
        logger.info(f"Would insert: '{document}' into '{collection.name}'.")


def migrate_orgs(
    client: MongoClient, old_db_name: str, new_db_name: str, dry_run: bool
):
    old_db = client[old_db_name]
    new_db = client[new_db_name]
    new_entity_col = new_db[EntityDAO.collection_name()]
    new_entity_col.create_indexes(EntityDAO.indexes())
    # Create root organization entity for '/' token
    root_org = EntityDAO(
        created_by=SYSTEM_INFOSTAR,
        handle="/",
        owner_ref=None,
        type="organization",
    )
    try:
        conditional_insert(new_entity_col, root_org.bson(), dry_run)
    except errors.DuplicateKeyError:
        pass
    obj = new_entity_col.find_one({"handle": "/"})
    if obj:
        root_org_ref = EntityDAO(**obj).to_ref()
    else:
        root_org_ref = EntityRef(handle="/", type="organization")
    # Query 'tokens' collection for all client names
    client_names = old_db["tokens"].distinct("client_name")
    for client_name in client_names:
        logger.debug(f"Creating organization entity for '{client_name}'")
        # Extract the root organization (e.g., `/teialabs/athena` -> `teialabs`)
        root_org_handle = client_name.split("/")[1]
        # Create organization entity based on client name
        org = EntityDAO(
            created_by=SYSTEM_INFOSTAR,
            handle=f"/{root_org_handle}",
            owner_ref=root_org_ref,
            type="organization",
        )
        try:
            conditional_insert(new_entity_col, org.bson(), dry_run)
        except errors.DuplicateKeyError:
            logger.warning(f"Organization entity '{org.handle}' already exists.")
    # query 'organizations' collection for all auth0 org-ids (/teialabs--auth0-org-id)
    orgs = old_db["organizations"].find()
    for org in orgs:
        conditional_action(
            Partial(
                new_entity_col.update_one,
                {"handle": org["name"]},
                {"$set": {"external_ids": org["external_ids"]}},
            ),
            dry_run,
        )


def migrate_services(
    client: MongoClient, old_db_name: str, new_db_name: str, dry_run: bool
):
    old_db = client[old_db_name]
    new_db = client[new_db_name]
    entity_col_new = new_db[EntityDAO.collection_name()]
    client_names = old_db["tokens"].distinct("client_name")
    # extract the service names (/teialabs/athena -> [athena], /osf/allai/code -> [allai, code]
    print(client_names)
    service_names = defaultdict(list)
    for client_name in client_names:
        org_name, *service_name = client_name.split("/")[1:]
        if service_name:
            service_names[f"/{org_name}"].append(service_name)
    for org_name, service_names in service_names.items():
        for serv_name in service_names:
            logger.debug(f"Creating service entity for '{org_name}' and '{serv_name}'")
            org = entity_col_new.find_one({"handle": org_name})
            if not org:
                raise ValueError(f"Organization entity '{org_name}' not found.")
            org_ref = EntityDAO(**org).to_ref()
            # create service entities based on service names
            service = EntityDAO(
                created_by=SYSTEM_INFOSTAR,
                handle=f"{'--'.join(serv_name)}",
                owner_ref=org_ref,
                type="service",
            )
            try:
                conditional_insert(entity_col_new, service.bson(), dry_run)
            except errors.DuplicateKeyError:
                pass


def migrate_users(
    client: MongoClient, old_db_name: str, new_db_name: str, dry_run: bool
):
    old_db = client[old_db_name]
    new_db = client[new_db_name]
    entity_col_new = new_db[EntityDAO.collection_name()]
    users = old_db["users"].find()
    for user in users:
        if "email" not in user:
            continue
        org_name = user["client_name"].split("/")[1]
        org = entity_col_new.find_one({"handle": f"/{org_name}"})
        if not org:
            raise ValueError(f"Organization entity '{org_name}' not found.")
        org_ref = EntityDAO(**org).to_ref()
        user_entity = EntityDAO(
            handle=user["email"],
            owner_ref=org_ref,
            type="user",
            extra=[
                Attribute(name="melt_key_client", value=user["client_name"]),
                Attribute(name="melt_key_client_first_login", value=user["created_at"]),
            ],
            created_by=Infostar(
                request_id=OID,
                apikey_name=user["created_by"]["token_name"],
                authprovider_org=org["handle"],
                authprovider_type="melt-key",
                extra={},
                service_handle="tauth",
                user_handle=user["created_by"]["user_email"],
                user_owner_handle=org["handle"],
                original=SYSTEM_INFOSTAR,
                client_ip=user["created_by"].get("user_ip", "127.0.0.1"),
            ),
        )
        try:
            conditional_action(
                Partial(entity_col_new.insert_one, user_entity.bson()), dry_run
            )
        except errors.DuplicateKeyError:
            logger.warning(f"User entity '{user_entity.handle}' already exists.")


def get_entity_ref_from_client_name(
    db: Database, client_name: str, mode: Literal["service", "organization"]
) -> EntityRef:
    entity_col = db[EntityDAO.collection_name()]
    org_name, *service_name = client_name.split("/")[1:]
    if mode == "service":
        entity = entity_col.find_one({"handle": "--".join(service_name)})
    elif mode == "organization":
        entity = entity_col.find_one({"handle": f"/{org_name}"})
    else:
        raise ValueError(f"Invalid mode '{mode}'.")
    if not entity:
        raise ValueError(f"Entity '{entity}' not found.")
    return EntityDAO(**entity).to_ref()


def migrate_melt_keys(
    client: MongoClient, old_db_name: str, new_db_name: str, dry_run: bool
):
    old_db = client[old_db_name]
    new_db = client[new_db_name]
    melt_keys = new_db[TokenDAO.collection_name()]
    tokens = list(old_db["tokens"].find())
    for t in tqdm(tokens):
        logger.debug(f"Creating melt-key for '{t['client_name']}'")
        org = get_entity_ref_from_client_name(new_db, t["client_name"], "organization")
        try:
            service_handle = get_entity_ref_from_client_name(
                new_db, t["client_name"], "service"
            ).handle
        except ValueError:
            service_handle = "tauth--api"
        key = TokenDAO(
            client_name=t["client_name"],
            name=t["name"],
            value=t["value"],
            created_by=Infostar(
                request_id=OID,
                apikey_name=t["created_by"]["token_name"],
                authprovider_org=org.handle,
                authprovider_type="melt-key",
                extra={},
                service_handle=service_handle,
                user_handle=t["created_by"]["user_email"],
                user_owner_handle=org.handle,
                original=SYSTEM_INFOSTAR,
                client_ip=t["created_by"].get("user_ip", "127.0.0.1"),
            ),
        )
        try:
            conditional_insert(melt_keys, key.bson(), dry_run)
        except errors.DuplicateKeyError:
            logger.warning(f"Melt-key '{key.name}' already exists.")


def migrate_authproviders(
    client: MongoClient, old_db_name: str, new_db_name: str, dry_run: bool
):
    old_db = client[old_db_name]
    new_db = client[new_db_name]
    authproviders = new_db[AuthProviderDAO.collection_name()]
    clients = old_db["clients"].distinct("name")
    orgs = [f"/{c.split('/')[1]}" for c in clients] + ["/"]
    for org in orgs:
        authprovider = AuthProviderDAO(
            created_by=SYSTEM_INFOSTAR,
            organization_ref=OrganizationRef(handle=org),
            service_ref=None,
            type="melt-key",
        )
        conditional_insert(authproviders, authprovider.bson(), dry_run)
    authproviders_data = [
        {
            "created_by": SYSTEM_INFOSTAR,
            "external_ids": [
                Attribute(
                    name="issuer", value="https://dev-z60iog20x0slfn0a.us.auth0.com"
                ),
                Attribute(name="audience", value="api://allai.chat.webui"),
                Attribute(name="client-id", value="4FdEO3ncOVFuROab8wf3c0GLyEMWi4f4"),
            ],
            "organization_ref": OrganizationRef(handle="/teialabs"),
            "service_ref": ServiceRef(handle="athena--chat"),
            "type": "auth0",
        },
        {
            "created_by": SYSTEM_INFOSTAR,
            "external_ids": [
                Attribute(name="issuer", value="https://osfdigital.eu.auth0.com"),
                Attribute(
                    name="audience", value="api://d4f7a0b5-2a6d-476f-b251-c468a25acdef"
                ),
                Attribute(name="client-id", value="My4fjEfyByLfRPWzxlkP7rTDSFroklNW"),
            ],
            "organization_ref": OrganizationRef(handle="/osf"),
            "service_ref": ServiceRef(handle="allai--chat"),
            "type": "auth0",
        },
        {
            "created_by": SYSTEM_INFOSTAR,
            "external_ids": [
                Attribute(name="issuer", value="https://osfdigital.eu.auth0.com"),
                Attribute(
                    name="audience", value="api://27a058b1-2f69-475b-b126-5d816b037cbe"
                ),
            ],
            "organization_ref": OrganizationRef(handle="/osf"),
            "service_ref": ServiceRef(handle="allai--code"),
            "type": "auth0",
        },
    ]
    for authprovider_data in authproviders_data:
        authprovider = AuthProviderDAO(**authprovider_data)
        conditional_insert(authproviders, authprovider.bson(), dry_run)


def main():
    args = get_args()
    logger.add("migration.log", level="INFO")
    logger.info("Starting migration script")
    client = MongoClient(args.host, args.port)
    # migrate_orgs(client, args.in_db, args.out_db, args.dry_run)
    # migrate_services(client, args.in_db, args.out_db, args.dry_run)
    # migrate_users(client, args.in_db, args.out_db, args.dry_run)
    migrate_melt_keys(client, args.in_db, args.out_db, args.dry_run)
    migrate_authproviders(client, args.in_db, args.out_db, args.dry_run)
    logger.info("Migration script completed")


if __name__ == "__main__":
    main()



from pathlib import Path

from fastapi import APIRouter, Body, Depends, HTTPException, Request
from fastapi import status as s
from loguru import logger

from ..authz.engines.factory import AuthorizationEngine
from ..dependencies.security import RequestAuthenticator
from ..entities.models import EntityDAO
from ..schemas import Infostar
from .policies.schemas import AuthorizationDataIn

service_name = Path(__file__).parent.name
router = APIRouter(prefix=f"/{service_name}", tags=[service_name + " 🔐"])


@router.post("", status_code=s.HTTP_200_OK)
@router.post("/", status_code=s.HTTP_200_OK, include_in_schema=False)
async def authorize(
    request: Request,
    authz_data: AuthorizationDataIn = Body(),
) -> dict:
    infostar = request.state.infostar
    logger.debug(f"Running authorization for user: {infostar}")
    logger.debug(f"Authorization data: {authz_data}")

    logger.debug("Getting authorization engine and adding context.")
    authz_engine = AuthorizationEngine.get()
    authz_data.context["infostar"] = infostar.model_dump(mode="json")
    authz_data.context["request"] = await request.json()
    entity = EntityDAO.from_handle(handle=infostar.user_handle)
    if not entity:
        message = f"Entity not found for handle: {infostar.user_handle}."
        logger.error(message)
        raise HTTPException(
            status_code=s.HTTP_401_UNAUTHORIZED,
            detail=dict(msg=message),
        )
    logger.debug(f"Entity found: {entity}.")

    logger.debug(f"Executing authorization logic.")
    # TODO: determine if we're gonna support arbitrary outputs here (e.g., filters)
    result = authz_engine.is_authorized(
        entity=entity,
        policy_name=authz_data.policy_name,
        resource=authz_data.resource,
        context=authz_data.context,
    )
    logger.debug(f"Authorization result: {result}.")
    result = dict(authorized=result)
    return result



from fastapi import APIRouter, Body, Depends, HTTPException, Path, Request
from fastapi import status as s
from http_error_schemas.schemas import RequestValidationError
from loguru import logger

from ..authn.melt_key.authorization import validate_scope_access_level
from ..authn.melt_key.models import TokenDAO
from ..authn.melt_key.schemas import TokenCreationIntermediate, TokenCreationOut
from ..authn.melt_key.token import create_token
from ..authproviders.models import AuthProviderDAO
from ..authz import privileges
from ..entities.models import EntityDAO
from ..schemas import Creator, Infostar
from ..settings import Settings
from ..utils import creation, reading
from .schemas import MeltAPIKeyIn

router = APIRouter(prefix="/keys", tags=["legacy"])


@router.post("", status_code=s.HTTP_201_CREATED)
@router.post("/", status_code=s.HTTP_201_CREATED, include_in_schema=False)
async def create_one(
    request: Request,
    body: MeltAPIKeyIn,
    infostar: Infostar = Depends(privileges.is_valid_user),
) -> TokenCreationOut:
    """
    Create a token.

    Clients can create tokens for themselves and their subclients, but not for parent clients.
    """
    creator: Creator = request.state.creator
    logger.debug(
        f"Attempting to create token {body.name!r} for {body.organization_handle!r}."
    )
    try:
        logger.debug(
            f"Checking if organization entity {body.organization_handle!r} exists."
        )
        filters = {"handle": body.organization_handle, "type": "organization"}
        org = reading.read_one_filters(infostar, model=EntityDAO, **filters)
    except HTTPException as e:
        details = RequestValidationError(
            loc=["path", "client_name"],
            msg="Cannot create token for non-existent organization entity.",
            type="DocumentNotFound",
        )
        raise HTTPException(status_code=s.HTTP_404_NOT_FOUND, detail=details)

    try:
        logger.debug(f"Checking if org {org.handle!r} has 'melt-key' authprovider.")
        filters = {"type": "melt-key", "organization_ref.handle": org.handle}
        _ = reading.read_one_filters(infostar, model=AuthProviderDAO, **filters)
    except HTTPException as e:
        details = RequestValidationError(
            loc=["path", "client_name"],
            msg="Cannot create token for organization with no 'melt-key' authprovider.",
            type="DocumentNotFound",
        )
        raise HTTPException(status_code=s.HTTP_404_NOT_FOUND, detail=details)

    logger.debug(
        f"Validating access: {infostar.authprovider_org!r} -> {body.organization_handle!r}."
    )
    validate_scope_access_level(body.organization_handle, infostar.authprovider_org)

    try:
        logger.debug(f"Creating {body.name!r} token.")
        token = TokenCreationIntermediate(
            client_name=creator.client_name,
            name=body.name,
            value=create_token(creator.client_name, body.name),
        )
        token = creation.create_one(token, model=TokenDAO, infostar=infostar)
        token_out = TokenCreationOut(
            **token.model_dump(exclude={"created_by"}),
            created_by=infostar,
        )
    except HTTPException as e:
        details = RequestValidationError(
            loc=["path", "name"],
            msg=f"Token {body.name!r} already exists.",
            type="DuplicateKeyError",
        )
        raise HTTPException(status_code=s.HTTP_409_CONFLICT, detail=details)

    return token_out


@router.delete(
    "/{client_name:path}/tokens/{token_name}", status_code=s.HTTP_204_NO_CONTENT
)
async def delete_one(
    request: Request,
    client_name: str = Path(),
    token_name: str = Path(),
    infostar: Infostar = Depends(privileges.is_valid_admin),
):
    """
    Delete a token.

    Clients can delete tokens for themselves and their subclients, but not for parent clients.
    """
    creator: Creator = request.state.creator
    logger.debug(f"Attempting to DELETE token {token_name!r} for {client_name!r}.")
    creator: Creator = request.state.creator
    try:
        org_handle = client_name.split("/")[1]
        organization = f"/{org_handle}"
        logger.debug(f"Checking if organization entity {organization!r} exists.")
        filters = {"handle": organization, "type": "organization"}
        reading.read_one_filters(infostar, model=EntityDAO, **filters)
    except HTTPException as e:
        details = RequestValidationError(
            loc=["path", "client_name"],
            msg="Cannot delete token for non-existent organization entity.",
            type="DocumentNotFound",
        )
        raise HTTPException(status_code=s.HTTP_404_NOT_FOUND, detail=details)

    logger.debug(f"Validating access: {creator.client_name!r} -> {client_name!r}.")
    validate_scope_access_level(client_name, creator.client_name)

    filters = {"client_name": client_name, "name": token_name}
    try:
        reading.read_one_filters(infostar, model=TokenDAO, **filters)
    except HTTPException as e:
        details = RequestValidationError(
            loc=["path", "token_name"],
            msg=f"Token {token_name!r} does not exist.",
            type="DocumentNotFound",
        )
        raise HTTPException(status_code=s.HTTP_404_NOT_FOUND, detail=details)

    logger.debug("Deleting token.")
    # TODO: needs soft delete.
    TokenDAO.collection(alias=Settings.get().TAUTH_REDBABY_ALIAS).delete_one(filters)



"""
Download all data from a MongoDB collection and save it as .csv for inspection.
"""
import argparse
import csv
import os
import sys

from pymongo import MongoClient


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--db", default="test", help="MongoDB database.")
    parser.add_argument("--collection", help="MongoDB collection.", required=True)
    parser.add_argument("--output", default="data.csv", help="Output file.")
    parser.add_argument(
        "--limit",
        default=0,
        type=int,
        help="Limit the number of documents to export.",
    )
    args = parser.parse_args()
    return args


def flatten(document: dict) -> dict:
    """
    Flatten a nested dictionary.

    >>> flatten({'a': {'b': 1, 'c': 2}})
    {'a.b': 1, 'a.c': 2}
    """
    flattened = {}
    for key, value in document.items():
        if isinstance(value, dict):
            for subkey, subvalue in flatten(value).items():
                flattened[f"{key}.{subkey}"] = subvalue
        else:
            flattened[key] = value
    return flattened


def main():
    args = get_args()
    client = MongoClient(args.host, args.port)
    db = client[args.db]
    collection = db[args.collection]
    cursor = collection.find()
    if args.limit > 0:
        cursor = cursor.limit(args.limit)
    first = collection.find({}).limit(1)[0]
    with open(args.output, "w") as csvfile:
        writer = csv.DictWriter(csvfile, fieldnames=flatten(first).keys())
        writer.writeheader()
        for document in cursor:
            writer.writerow(flatten(document))


if __name__ == "__main__":
    main()



from datetime import datetime

from pydantic import BaseModel, EmailStr, Field

from ...schemas import Creator, Infostar


class TokenMeta(BaseModel):
    created_at: datetime
    created_by: Infostar
    name: str


class TokenCreationIntermediate(BaseModel):
    client_name: str
    name: str
    value: str


class TokenCreationOut(TokenMeta):
    value: str


class UserCreation(BaseModel):
    email: EmailStr


class UserOut(BaseModel):
    created_at: datetime
    created_by: Creator
    email: EmailStr


class ClientCreation(BaseModel):
    name: str


class ClientCreationOut(BaseModel):
    created_at: datetime
    created_by: Creator
    id: str = Field(alias="_id")
    name: str
    tokens: list[TokenCreationOut]
    users: list[UserOut]


class ClientOut(BaseModel):
    created_at: datetime
    created_by: Creator
    id: str = Field(alias="_id")
    name: str


class ClientOutJoinTokensAndUsers(BaseModel):
    created_at: datetime
    created_by: Creator
    id: str = Field(alias="_id")
    name: str
    tokens: list[TokenMeta]
    users: list[UserOut]



from typing import Optional, TypedDict

from pydantic import BaseModel, EmailStr
from redbaby.pyobjectid import PyObjectId


class InfostarExtra(TypedDict, total=False):
    geolocation: str
    jwt_sub: str
    os: str
    url: str  # reverse dns
    user_agent: str


class Infostar(BaseModel):
    request_id: PyObjectId  # will be propagated across all services
    apikey_name: str = "jwt"  # specific api key name (nei.workstation.homeoffice)
    authprovider_type: str  # auth0
    authprovider_org: str  # /teialabs
    extra: InfostarExtra
    service_handle: str  # e.g., allai--code
    user_handle: EmailStr  # email
    user_owner_handle: str  # e.g., organization, user family, ...
    client_ip: str
    original: Optional["Infostar"] = None  # if any attributes were overriden



from typing import Literal, Optional

from pydantic import BaseModel, Field
from pymongo import IndexModel
from redbaby.behaviors.hashids import HashIdMixin
from redbaby.behaviors.reading import ReadingMixin
from redbaby.document import Document

from ..schemas.attribute import Attribute
from ..utils.teia_behaviors import Authoring


class EntityRefBase(BaseModel):
    handle: str


class EntityRef(EntityRefBase):
    type: Literal["organization", "service", "user"]


class OrganizationRef(EntityRefBase):
    type: Literal["organization"] = "organization"


class ServiceRef(EntityRefBase):
    type: Literal["service"] = "service"


class UserRef(EntityRefBase):
    type: Literal["user"] = "user"


class EntityDAO(Document, Authoring, ReadingMixin, HashIdMixin):
    external_ids: list[Attribute] = Field(
        default_factory=list
    )  # e.g., url, azuread-id/auth0-id, ...
    extra: list[Attribute] = Field(default_factory=list)
    handle: str
    owner_ref: Optional[EntityRef]
    roles: list[str] = Field(
        default_factory=list
    )  # e.g.: ["teia-admin", "allai-user-basic"]
    type: Literal["user", "service", "organization"]

    @classmethod
    def collection_name(cls) -> str:
        return "entities"

    @classmethod
    def indexes(cls) -> list[IndexModel]:
        idxs = [
            IndexModel("roles"),
            IndexModel(
                [("type", 1), ("handle", 1), ("owner_ref.handle", 1)], unique=True
            ),
            IndexModel(
                [("type", 1), ("external_ids.name", 1), ("external_ids.value", 1)],
            ),
        ]
        return idxs

    @classmethod
    def from_handle(cls, handle: str) -> Optional["EntityDAO"]:
        out = cls.collection(alias="tauth").find_one({"handle": handle})
        if out:
            return EntityDAO(**out)

    @classmethod
    def from_handle_to_ref(cls, handle: str) -> Optional[EntityRef]:
        entity = cls.from_handle(handle)
        if entity:
            return EntityRef(type=entity.type, handle=entity.handle, id=entity.id)

    def to_ref(self) -> EntityRef:
        return EntityRef(type=self.type, handle=self.handle, id=self.id)

    def hashable_fields(self) -> list[str]:
        l = [self.handle]
        if self.owner_ref:
            l.append(self.owner_ref.handle)
        return l



from typing import Iterable

from fastapi import FastAPI, Header, HTTPException, Request, Security
from fastapi.security.http import HTTPAuthorizationCredentials, HTTPBase
from http_error_schemas.schemas import RequestValidationError
from loguru import logger

from ..authn import auth0_dyn
from ..authn.melt_key import authentication as melt_key


def init_app(app: FastAPI):
    from tauth.utils.fastapi_extension import get_depends

    app.router.dependencies.append(get_depends())


class RequestAuthenticator:
    ignore_paths: Iterable[str] = ("/", "/api", "/api/")

    @staticmethod
    def validate(
        request: Request,
        user_email: str | None = Header(
            default=None, alias="X-User-Email", description="Ignore when using OAuth."
        ),
        id_token: str | None = Header(
            default=None, alias="X-ID-Token", description="Auth0 ID token."
        ),
        authorization: HTTPAuthorizationCredentials | None = Security(
            HTTPBase(scheme="bearer", auto_error=False)
        ),
    ):
        # TODO: move empty header check to a subclass of HTTPBase
        req_path: str = request.scope["path"]
        if request.method == "GET" and req_path in RequestAuthenticator.ignore_paths:
            return

        if not authorization:
            d = RequestValidationError(
                loc=["header", "Authorization"],
                msg="Missing Authorization header.",
                type="MissingHeader",
            )
            raise HTTPException(401, detail=d)

        token_type, token_value = authorization.scheme, authorization.credentials
        if token_type.lower() != "bearer":
            raise HTTPException(
                401, detail={"msg": "Invalid authorization scheme; expected 'bearer'."}
            )

        if token_value.startswith("MELT_"):
            logger.debug("Authenticating with a MELT API key (legacy).")
            melt_key.RequestAuthenticator.validate(request, user_email, token_value)
            return
        if id_token is None:
            raise HTTPException(401, detail={"msg": "Missing ID token."})
        else:
            logger.debug("Authenticating with an Auth0 provider.")
            # figure out which provider/iss it's from
            auth0_dyn.RequestAuthenticator.validate(request, token_value, id_token)
            return
        # TODO: check if it starts with TAUTH_
        # TODO: check if it's a JWT

        raise HTTPException(401, detail={"msg": "No authentication method succeeded."})



from pathlib import Path
from typing import Optional

from fastapi import APIRouter, Body, Depends, Query, Request
from fastapi import status as s

from ..authz import privileges
from ..schemas import Infostar
from ..schemas.gen_fields import GeneratedFields
from ..utils import creation, reading
from .models import EntityDAO
from .schemas import EntityIn, EntityIntermediate

service_name = Path(__file__).parent.name
router = APIRouter(prefix=f"/{service_name}", tags=[service_name + " 👥💻🏢"])


@router.post("", status_code=s.HTTP_201_CREATED)
@router.post("/", status_code=s.HTTP_201_CREATED, include_in_schema=False)
async def create_one(
    request: Request,
    body: EntityIn = Body(
        openapi_examples=EntityIn.model_config["json_schema_extra"]["examples"][0]
    ),
    infostar: Infostar = Depends(privileges.is_valid_admin),
):
    if body.owner_handle:
        owner_ref = EntityDAO.from_handle_to_ref(body.owner_handle)
    else:
        owner_ref = None
    schema_in = EntityIntermediate(owner_ref=owner_ref, **body.model_dump())
    entity = creation.create_one(schema_in, EntityDAO, infostar)
    return GeneratedFields(**entity.model_dump(by_alias=True))


@router.get("", status_code=s.HTTP_200_OK)
@router.get("/", status_code=s.HTTP_200_OK, include_in_schema=False)
async def read_many(
    request: Request,
    infostar: Infostar = Depends(privileges.is_valid_user),
    name: Optional[str] = Query(None),
    external_id_key: Optional[str] = Query(None, alias="external_ids.key"),
    external_id_value: Optional[str] = Query(None, alias="external_ids.value"),
):
    orgs = reading.read_many(
        infostar=infostar,
        model=EntityDAO,
        **request.query_params,
    )
    return orgs



from typing import Self

from pydantic import BaseModel, EmailStr

from .infostar import Infostar


class Creator(BaseModel):
    client_name: str
    token_name: str
    user_email: EmailStr
    user_ip: str = "127.0.0.1"

    @classmethod
    def from_infostar(cls, infostar: Infostar) -> Self:
        if not infostar.service_handle:
            client_name = infostar.user_owner_handle
        else:
            client_name = f"{infostar.user_owner_handle}/{infostar.service_handle}"
        c = cls(
            client_name=client_name,
            token_name=infostar.apikey_name,
            user_email=infostar.user_handle,
            user_ip=infostar.client_ip,
        )
        return c



import time
from collections import deque
from typing import Any, TypedDict

from cryptography.hazmat.primitives.asymmetric import rsa

from .constants import DEFAULT_ROTATION_LENGTH, DEFAULT_ROTATION_TTL
from .errors import InvalidToken
from .keygen import JWKS, generate_jwks, generate_signing_pair
from .tokcoder import Claims, decode_jwt, encode_jwt, unverified_headers


class KeyPair(TypedDict):
    private: rsa.RSAPrivateKey
    public: rsa.RSAPublicKey
    jwks: JWKS
    ttl: float


def init_key_pair() -> KeyPair:
    private, public = generate_signing_pair()
    return {
        "private": private,
        "public": public,
        "jwks": generate_jwks(public),
        "ttl": time.time(),
    }


class CredentialStore:
    _rotations: deque[KeyPair] = deque([], maxlen=DEFAULT_ROTATION_LENGTH)
    _rotation_ttl: float = DEFAULT_ROTATION_TTL
    _current: KeyPair = init_key_pair()

    @classmethod
    def setup(
        cls,
        rotation_length: int | None = None,
        rotation_ttl: int | None = None,
        key_pair: KeyPair | None = None,
    ):
        if rotation_length is not None:
            cls._rotations = deque([], maxlen=rotation_length)

        if rotation_ttl is not None:
            cls._rotation_ttl = rotation_ttl

        if key_pair is not None:
            cls._current = key_pair

    @classmethod
    def rotate(cls):
        private, public = generate_signing_pair()
        cls._rotations.append(cls._current)
        cls._current = {
            "private": private,
            "public": public,
            "jwks": generate_jwks(public),
            "ttl": time.time(),
        }
        return cls._current

    @classmethod
    def pair(cls):
        return cls._current

    @classmethod
    def public(cls):
        return cls._current["public"]

    @classmethod
    def private(cls):
        return cls._current["private"]

    @classmethod
    def jwks(cls) -> JWKS:
        return cls._current["jwks"]

    @classmethod
    def alive(cls):
        return cls._current["ttl"] + DEFAULT_ROTATION_TTL < time.time()

    @classmethod
    def rotations(cls):
        return deque([cls._current]) + cls._rotations

    @classmethod
    def encode(cls, payload: dict[str, Any]) -> bytes:
        if not cls.alive():
            cls.rotate()
        return encode_jwt(payload=payload, private_key=cls.private())

    @classmethod
    def decode(cls, token: str | bytes, claims: Claims) -> dict[str, Any]:
        if isinstance(token, str):
            token = token.encode("utf-8")

        e = "Invalid token."
        for rotation in cls.rotations():
            try:
                decoded = decode_jwt(
                    token=token,
                    public_key=rotation["public"],
                    claims=claims,
                )
                return decoded
            except InvalidToken as err:
                e = str(err)

        raise InvalidToken(e)

    @classmethod
    def headers(cls, token: str | bytes) -> dict[str, Any]:
        if isinstance(token, str):
            token = token.encode("utf-8")

        return unverified_headers(token=token)



import secrets
from typing import Optional

from fastapi import HTTPException, Request
from fastapi import status as s
from loguru import logger
from pydantic import validate_email
from redbaby.pyobjectid import PyObjectId

from ...authproviders.models import AuthProviderDAO
from ...authproviders.schemas import AuthProviderMoreIn
from ...entities.models import EntityDAO, EntityRef, OrganizationRef
from ...entities.schemas import EntityIntermediate
from ...schemas import Creator, Infostar
from ...schemas.attribute import Attribute
from ...settings import Settings
from ...utils import creation, reading
from .token import parse_token, sanitize_client_name, validate_token_against_db

EmailStr = str


class RequestAuthenticator:
    @classmethod
    def validate(
        cls,
        request: Request,
        user_email: str | None,
        api_key_header: str,
    ):
        creator, token_creator_user_email = cls.get_request_creator(
            api_key_header, user_email
        )
        infostar = cls.get_request_infostar(creator, request)
        # TODO: create user entity in DB
        cls.create_user_on_db(creator, infostar, token_creator_user_email)

        if request.client is not None:
            creator.user_ip = request.client.host
        if request.headers.get("x-forwarded-for"):
            creator.user_ip = request.headers["x-forwarded-for"]
        request.state.creator = creator
        request.state.infostar = infostar

    @staticmethod
    def get_token_details(token_value: str) -> tuple[str, str, str]:
        """Returns the client and token name from the token value."""
        logger.debug("Parsing token to extract details.")
        client, name, value = parse_token(token_value)
        client = sanitize_client_name(client)
        logger.debug(f"client_name: {client!r}, token_name: {name!r}")
        return client, name, value

    @staticmethod
    def get_organization_for_client(creator: Creator, infostar: Infostar) -> EntityDAO:
        """Return the organization entity for a given client name."""
        if creator.client_name == "/":
            logger.debug(f"Using root token, org = /.")
            organization = "/"
            org_i = EntityIntermediate(
                handle="/",
                owner_ref=None,
                type="organization",
            )
            try:  # TODO read or create
                org = creation.create_one(org_i, EntityDAO, infostar=infostar)
                logger.debug(f"Created root organization entity.")
            except HTTPException as e:
                if e.status_code != 409:
                    raise e
        logger.debug(f"Getting organization for client {creator.client_name!r}.")
        org_handle = creator.client_name.split("/")[1]
        organization = f"/{org_handle}"
        filters_org = {"type": "organization", "handle": organization}
        org = reading.read_one_filters(
            infostar=infostar, model=EntityDAO, **filters_org
        )
        return org

    @staticmethod
    def get_authprovider(
        organization: EntityDAO, creator: Creator, infostar: Infostar
    ) -> AuthProviderDAO:
        """Return the 'melt-key' AuthProvider for the given client name."""
        logger.debug(f"Getting 'melt-key' provider for org {organization.handle!r}.")
        if organization.handle == "/":
            provider_i = AuthProviderMoreIn(
                organization_ref=OrganizationRef(**organization.model_dump()),
                type="melt-key",
                service_ref=None,
                extra=[Attribute(name="melt_key_client", value="/")],
            )
            try:  # TODO read or create
                provider = creation.create_one(
                    provider_i, AuthProviderDAO, infostar=infostar
                )
                logger.debug(
                    f"Created 'melt-key' provider for org {organization.handle!r}."
                )
            except HTTPException as e:
                if e.status_code != 409:
                    raise e
        filters = {
            "type": "melt-key",
            "organization_ref.handle": organization.handle,
        }
        provider = reading.read_one_filters(infostar, model=AuthProviderDAO, **filters)
        return provider

    @staticmethod
    def get_request_infostar(creator: Creator, request: Request):
        logger.debug("Assembling Infostar based on Creator.")
        breadcrumbs = creator.client_name.split("/")
        owner_handle = f"/{breadcrumbs[1]}"
        service_handle = "--".join(breadcrumbs[2:]) if len(breadcrumbs) > 2 else ""
        infostar = Infostar(
            request_id=PyObjectId(),
            apikey_name=creator.token_name,
            authprovider_type="melt-key",
            authprovider_org=owner_handle,
            # extra=InfostarExtra(
            #     geolocation=request.headers.get("x-geo-location"),
            #     jwt_sub=request.headers.get("x-jwt-sub"),
            #     os=request.headers.get("x-os"),
            #     url=request.headers.get("x-url"),
            #     user_agent=request.headers.get("user-agent"),
            # ),
            extra={},
            original=None,
            service_handle=service_handle,
            user_handle=creator.user_email,
            user_owner_handle=owner_handle,
            client_ip=creator.user_ip,
        )
        return infostar

    @classmethod
    def get_request_creator(
        cls, token: str, user_email: Optional[str]
    ) -> tuple[Creator, Optional[EmailStr]]:
        """Returns the Creator and token creator user email for a given request."""
        logger.debug("Getting Creator for request.")
        client_name, token_name, token_value = cls.get_token_details(token)
        token_creator_user_email = None
        if client_name == "/":
            logger.debug("Using root token, checking email.")
            if user_email is None:
                code, m = (
                    s.HTTP_401_UNAUTHORIZED,
                    "User email is required for root client.",
                )
                raise HTTPException(status_code=code, detail=m)
            if not secrets.compare_digest(token, Settings().TAUTH_ROOT_API_KEY):
                print(token_value, Settings().TAUTH_ROOT_API_KEY)
                code, m = s.HTTP_401_UNAUTHORIZED, "Root token does not match env var."
                raise HTTPException(status_code=code, detail=m)
            try:
                validate_email(user_email)
            except:
                code, m = s.HTTP_401_UNAUTHORIZED, "User email is not valid."
                raise HTTPException(status_code=code, detail=m)
            request_creator_user_email = user_email
        else:
            logger.debug("Using non-root token, validating token in DB.")
            token_obj = validate_token_against_db(token, client_name, token_name)
            if user_email is None:
                request_creator_user_email = token_obj["created_by"]["user_handle"]
            else:
                token_creator_user_email = token_obj["created_by"]["user_handle"]
                try:
                    validate_email(user_email)
                except:
                    code, m = s.HTTP_401_UNAUTHORIZED, "User email is not valid."
                    raise HTTPException(status_code=code, detail=m)
                request_creator_user_email = user_email

        creator = Creator(
            client_name=client_name,
            token_name=token_name,
            user_email=request_creator_user_email,
        )
        return creator, token_creator_user_email

    @classmethod
    def create_user_on_db(
        cls,
        creator: Creator,
        infostar: Infostar,
        token_creator_email: Optional[EmailStr],
    ):
        logger.debug("Registering user.")
        client_org = cls.get_organization_for_client(creator, infostar)
        authprovider = cls.get_authprovider(client_org, creator, infostar)
        org_ref = EntityRef(**client_org.model_dump())
        melt_key_client_extra = Attribute(
            name="melt_key_client", value=creator.client_name
        )
        user_creator_email = (
            creator.user_email if token_creator_email is None else token_creator_email
        )

        user_data = None
        try:
            logger.debug(
                f"Checking if user {user_creator_email!r} exists in org {client_org.handle!r}."
            )
            user_data = reading.read_one_filters(
                infostar,
                model=EntityDAO,
                handle=user_creator_email,
                owner_ref=org_ref.model_dump(),
                type="user",
            )
        except HTTPException as e:
            if e.status_code != 404:
                raise e

        if user_data:
            logger.debug(
                f"User {user_creator_email!r} exists. Adding {creator.client_name!r} client info."
            )
            entity_coll = EntityDAO.collection(alias=Settings.get().TAUTH_REDBABY_ALIAS)
            entity_coll.update_one(
                filter={"_id": user_data.id},
                update={"$addToSet": {"extra": melt_key_client_extra.model_dump()}},
            )
            return

        logger.debug(f"User does not exist. Adding user to DB.")
        user_i = EntityIntermediate(
            handle=creator.user_email,
            owner_ref=org_ref,
            type="user",
            extra=[melt_key_client_extra],
        )
        user = creation.create_one(user_i, EntityDAO, infostar=infostar)
        logger.debug(f"Registered {user_creator_email!r} in org {client_org.handle!r}.")



from typing import Any, Callable, Generic, TypedDict, TypeVar

from authlib.common.encoding import to_bytes
from authlib.jose import JWTClaims, jwt, util
from authlib.jose.errors import BadSignatureError, DecodeError, InvalidClaimError
from cryptography.hazmat.primitives.asymmetric import rsa

from .errors import InvalidToken
from .keygen import private_key_bytes, public_key_bytes

T = TypeVar("T")


class ClaimValue(Generic[T], TypedDict, total=False):
    value: T
    values: list[T]
    validate: Callable[[JWTClaims, T], bool]


class Claims(TypedDict, total=False):
    iss: ClaimValue[str]
    sub: ClaimValue[str]
    aud: ClaimValue[str]
    exp: ClaimValue[float]
    nbf: ClaimValue[float]
    iat: ClaimValue[float]
    jti: ClaimValue[str]


def encode_jwt(payload: dict[str, Any], private_key: rsa.RSAPrivateKey) -> bytes:
    """
    Encodes a JWT token with the given payload and private key.
    """

    header = {"alg": "RS256"}
    key_bytes = private_key_bytes(private_key)
    return jwt.encode(header=header, payload=payload, key=key_bytes)


def decode_jwt(
    token: bytes,
    public_key: rsa.RSAPublicKey,
    claims: Claims,
) -> dict[str, Any]:
    """
    Decodes a JWT token with the given public key.
    """

    key_bytes = public_key_bytes(public_key)
    try:
        c = jwt.decode(token, key=key_bytes, claims_options=claims)
    except BadSignatureError:
        raise InvalidToken("Invalid signature.")

    if not claims:
        return c["payload"]

    try:
        c.validate()
    except InvalidClaimError:
        raise InvalidToken("Invalid token claims.")

    return c["payload"]


def unverified_headers(token: bytes):
    try:
        signing_input, _ = token.rsplit(b".", 1)
        protected_segment, _ = signing_input.split(b".", 1)
    except ValueError:
        raise DecodeError("Not enough segments")

    protected = util.extract_header(protected_segment, DecodeError)
    return protected



from typing import Type, TypeVar

from fastapi import HTTPException
from pydantic import BaseModel
from pymongo import errors as pymongo_errors
from redbaby.document import Document

from ..schemas import Infostar
from ..settings import Settings

T = TypeVar("T", bound=Document)


def create_one(item_in: BaseModel, model: Type[T], infostar: Infostar) -> T:
    item = model(**item_in.model_dump(), created_by=infostar)  # type: ignore
    try:
        res = model.collection(alias=Settings.get().TAUTH_REDBABY_ALIAS).insert_one(
            item.bson()
        )
    except pymongo_errors.DuplicateKeyError as e:
        d = {
            "error": e.__class__.__name__,
            "msg": e._message,
            "details": e.details,
        }
        raise HTTPException(status_code=409, detail=d)
    # TODO: add logging
    # TODO: add event tracking
    return item
